@@ -1,8 +1,5 @@
 #!/usr/bin/env python
 
-from __future__ import print_function
-from __future__ import absolute_import
-
 import os
 import time
 import itertools
@@ -15,12 +12,14 @@ import matplotlib as mpl
 mpl.use('Agg')
 
 import matplotlib.pyplot as plt
-
 import astropy.constants as const
 
-# All ps_eor modules are lazy loaded
 from ps_eor import __version__
 
+# lazy loaded:
+# All ps_eor modules
+# astropy.units as u in gen_vis_cube
+
 
 mpl.rcParams['image.cmap'] = 'Spectral_r'
 mpl.rcParams['image.origin'] = 'lower'
@@ -30,8 +29,6 @@ mpl.rcParams['axes.grid'] = True
 t_file = click.Path(exists=True, dir_okay=False)
 t_dir = click.Path(exists=True, file_okay=False)
 
-INSTRU_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'instru')
-
 MIN_EOR_BW_MHZ = 2
 
 
@@ -356,6 +353,22 @@ def do_gpr(i_cube, v_cube, gpr_config_i, gpr_config_v, rnd_seed=1):
     return gpr_res_i, gpr_res_v
 
 
+def do_ml_gpr(i_cube, noise_cube, ml_gpr_config, save_dir='.', save_name=None, rnd_seed=1):
+    from ps_eor import ml_gpr
+
+    # Saving PS stacker results for the residual, the noie
+    print('Running ML-GPR for Stokes I ...\n')
+    t = time.time()
+
+    fitter = ml_gpr.MLGPRForegroundFitter(ml_gpr_config)
+
+    ml_gpr_res = fitter.run(i_cube, noise_cube, save_dir=save_dir, save_name=save_name, verbose=True)
+
+    print(f"\nDone in {(time.time() - t):.2f} second\n")
+
+    return ml_gpr_res
+
+
 def plot_img_and_freq_slice(eor, data_cube, ax1, ax2, name):
     img_cube = eor.get_slice(data_cube).regrid().image()
 
@@ -580,6 +593,200 @@ def run_gpr(file_i, file_v, gpr_config_i, gpr_config_v, flag_config, eor_bins_li
         fig.savefig(os.path.join(plot_out_dir, 'ps3d.pdf'), bbox_inches='tight')
 
 
+@main.command('run_ml_gpr')
+@click.argument('file_i', type=t_file)
+@click.argument('file_v', type=t_file)
+@click.argument('ml_gpr_config_file', type=t_file)
+@click.option('--flag_config', '-f', type=t_file, help='GPR configuration parset for Stokes I')
+@click.option('--eor_bins_list', '-e', type=t_file, help='Listing of EoR redshift bins')
+@click.option('--ps_conf', '-c', type=t_file, help='Power spectra configuration file')
+@click.option('--output_dir', '-o', help='Output directory', default='.')
+@click.option('--plots_output_dir', '-po', help='Output directory for the figures')
+@click.option('--no_plot', help='Skip plotting', is_flag=True)
+@click.option('--rnd_seed', help='Set a random seed', default=3, type=int)
+@click.option('--file_noise', help='Input noise datacube. If not set, Stokes V frequency diff will be used.',
+              type=t_file)
+def run_ml_gpr(file_i, file_v, ml_gpr_config_file, flag_config, eor_bins_list, ps_conf,
+               output_dir, plots_output_dir, no_plot, rnd_seed, file_noise):
+    ''' Run GPR & generate power spectra
+
+    \b
+    FILE_I: Input Stoke I datacube
+    FILE_V: Input Stoke V datacube
+    GPR_CONFIG_I: GPR configuration parset for Stokes I
+    '''
+    from ps_eor import datacube, pspec, flagger, psutil, ml_gpr
+
+    if plots_output_dir is None:
+        plots_output_dir = output_dir
+
+    if not os.path.exists(output_dir):
+        os.makedirs(output_dir)
+
+    if not os.path.exists(plots_output_dir):
+        os.makedirs(plots_output_dir)
+
+    print(f'Loading configuration file {ml_gpr_config_file} ...')
+    ml_gpr_config = ml_gpr.MLGPRConfigFile.load_with_defaults(ml_gpr_config_file, check_args=False)
+
+    print('Loading data ...')
+    i_cube = datacube.DataCube.load(file_i)
+    v_cube = datacube.DataCube.load(file_v)
+
+    ps_builder = pspec.PowerSpectraBuilder(ps_conf, eor_bins_list)
+    ps_conf = ps_builder.ps_config
+
+    i_cube.filter_uvrange(ps_conf.umin, ps_conf.umax)
+    v_cube.filter_uvrange(ps_conf.umin, ps_conf.umax)
+
+    if flag_config:
+        print('Running flagger ...')
+        flagger_runner = flagger.FlaggerRunner.load(flag_config)
+        i_cube, v_cube = flagger_runner.run(i_cube, v_cube)
+
+        if not no_plot:
+            fig = flagger_runner.plot()
+            fig.savefig(os.path.join(plots_output_dir, 'flagger_gpr.pdf'))
+
+    if file_noise is not None:
+        noise_cube = datacube.DataCube.load(file_noise)
+    else:
+        if ml_gpr_config.kern.noise.estimate_baseline_noise_from_stokes == 'V':
+            noise_cube_df = v_cube.make_diff_cube()
+        elif ml_gpr_config.kern.noise.estimate_baseline_noise_from_stokes == 'I':
+            noise_cube_df = i_cube.make_diff_cube()
+        else:
+            print('Error: estimate_baseline_noise_from_stokes must be I or V')
+            return
+
+        w = i_cube.weights.copy()
+        w.scale_with_noise_cube(noise_cube_df, ps_conf.empirical_weighting_polyfit_deg)
+        sefd = v_cube.make_diff_cube().estimate_sefd()
+        print(f'Using simulated noise with SEFD={sefd:.1f} Jy')
+        noise_cube = w.simulate_noise(sefd, i_cube.meta.total_time, fake_apply_win_fct=True)
+
+    ps_conf.el = 2 * np.pi * np.arange(i_cube.ru.min(), i_cube.ru.max(), ps_conf.du)
+
+    k_fg, k_eor, k_noise = ml_gpr_config.get_kern()
+
+    print(f'\nKern FG\n{k_fg}')
+    print(f'Kern 21-cm\n{k_eor}')
+    print(f'Kern noise\n{k_noise}')
+
+    for eor_name in ps_builder.eor_bin_list.get_all_names():
+        eor = ps_builder.eor_bin_list.get(eor_name, freqs=i_cube.freqs)
+
+        if not eor or eor.bw_total < MIN_EOR_BW_MHZ * 1e6:
+            continue
+
+        ps_gen = ps_builder.get(i_cube, eor_name)
+
+        out_dir = os.path.join(output_dir, 'eor%s_u%s-%s' % (eor_name, int(ps_conf.umin), int(ps_conf.umax)))
+        plot_out_dir = os.path.join(plots_output_dir, 'eor%s_u%s-%s' %
+                                    (eor_name, int(ps_conf.umin), int(ps_conf.umax)))
+
+        if not os.path.exists(out_dir):
+            os.makedirs(out_dir)
+
+        if not os.path.exists(plot_out_dir):
+            os.makedirs(plot_out_dir)
+
+        ps_gen_fg = ps_builder.get(i_cube, eor_name, window_fct='hann', rmean_freqs=True)
+
+        if ps_conf.empirical_weighting:
+            sefd = v_cube.make_diff_cube().estimate_sefd()
+            poly_deg = ps_conf.empirical_weighting_polyfit_deg
+            v_cube.weights.scale_with_noise_cube(eor.get_slice(
+                v_cube.make_diff_cube()), sefd_poly_fit_deg=poly_deg, expected_sefd=sefd)
+            i_cube.weights.scale_with_noise_cube(eor.get_slice(
+                v_cube.make_diff_cube()), sefd_poly_fit_deg=poly_deg, expected_sefd=sefd)
+
+        kbins = np.logspace(np.log10(ps_gen.kmin), np.log10(ps_conf.kbins_kmax), ps_conf.kbins_n)
+
+        print('\nRunning ML GPR for EoR bin:', eor.name)
+        print('Frequency range: %.2f - %.2f MHz (%i SB)\n' % (eor.fmhz[0], eor.fmhz[-1], len(eor.fmhz)))
+        print('Mean redshift: %.2f (%.2f MHz)\n' % (eor.z, eor.mfreq * 1e-6))
+
+        i_cube_slice = eor.get_slice_fg(i_cube)
+        n_cube_slice = eor.get_slice_fg(noise_cube)
+        ml_gpr_res = do_ml_gpr(i_cube_slice, n_cube_slice, ml_gpr_config, save_dir=out_dir,
+                               save_name='ml_gpr_res_I', rnd_seed=rnd_seed)
+
+        # Get results
+        scaled_noise_cube = ml_gpr_res.get_scaled_noise_cube()
+        ps_fg = ml_gpr_res.get_ps_fg(ps_gen_fg, kbins)
+        ps_eor = ml_gpr_res.get_ps_eor(ps_gen, kbins)
+        ps_residual = ml_gpr_res.get_ps_res(ps_gen, kbins)
+
+        # Saving 2D and 3D residuals I and noise PS
+        ratio, ratio_high, ratio_low = get_ratios(ps_gen, ps_residual.get_ps2d() / ps_gen.get_ps2d(scaled_noise_cube))
+        ratio_txt = 'Ratio I/noise (med: %.2f / %.2f / %.2f)' % (ratio, ratio_high, ratio_low)
+        print(ratio_txt)
+
+        ps_residual.get_ps2d().save(os.path.join(out_dir, 'ps2d_I_residual.mc.h5'))
+        ps_eor.get_ps2d().save_to_txt(os.path.join(out_dir, 'ps2d_eor.mc.h5'))
+        ps_fg.get_ps2d().save_to_txt(os.path.join(out_dir, 'ps2d_fg.mc.h5'))
+
+        ps3d_noise = ps_gen.get_ps3d(kbins, scaled_noise_cube)
+        ps3d_res = ps_residual.get_ps3d()
+        ps3d_rec = (ps3d_res - ps3d_noise)
+
+        ps3d_res.save_to_txt(os.path.join(out_dir, 'ps3d_I_residual.mc.h5'))
+        ps3d_rec.save(os.path.join(out_dir, 'ps3d_I_minus_noise.mc.h5'))
+        ps3d_noise.save_to_txt(os.path.join(out_dir, 'ps3d_I_noise.mc.h5'))
+
+        up = ps3d_rec.get_upper()
+        k_up = ps3d_rec.k_mean[np.nanargmin(up)]
+        upper_limit = '2 sigma upper limit: (%.1f)^2 mK^2 at k = %.3f' % (np.nanmin(up), k_up)
+        print(upper_limit)
+
+        if no_plot:
+            continue
+
+        print('Plotting results ...\n')
+
+        # Plotting MCMC results
+
+        ml_gpr_res.sampler.plot_corner(ml_gpr_config.mcmc.n_burn).savefig(
+            os.path.join(plot_out_dir, 'mcmc_corner.pdf'))
+        ml_gpr_res.sampler.plot_samples().savefig(os.path.join(plot_out_dir, 'mcmc_samples.pdf'))
+        ml_gpr_res.sampler.plot_samples_likelihood(n_burn=ml_gpr_config.mcmc.n_burn
+                                                   ).savefig(os.path.join(plot_out_dir, 'mcmc_samples_likelihood.pdf'))
+
+        # Plotting GPR results
+        save_ps2d(ps_gen, ps_residual.get_ps(), ps_residual.get_ps2d(), 'Stokes I residual',
+                  os.path.join(plot_out_dir, 'ps2d_I_residual.pdf'))
+
+        save_ps2d(ps_gen, ps_residual.get_ps() / ps_gen.get_ps(scaled_noise_cube),
+                  ps_residual.get_ps2d() / ps_gen.get_ps2d(scaled_noise_cube), 'I residual / noise',
+                  os.path.join(plot_out_dir, 'ps2d_I_over_noise.pdf'), vmin=0.5, vmax=10)
+
+        fp = pspec.FourPanelPsResults(ps_gen, kbins)
+        fp.add_cube(i_cube, 'I', ps_gen=ps_gen_fg, c=psutil.lblue)
+        fp.add_cube(v_cube, 'V', c=psutil.lgreen)
+        fp.add_cube(n_cube_slice, 'Input noise', c=psutil.red)
+        fp.add_ps_stacker(ps_residual, 'I residual', c=psutil.dblue)
+        fp.done(ncol_legend=4)
+        fp.savefig(os.path.join(plot_out_dir, 'ps_I_V_residuals.pdf'))
+
+        fp = pspec.FourPanelPsResults(ps_gen, kbins)
+        fp.add_cube(i_cube, 'I', ps_gen=ps_gen_fg, c=psutil.lblue)
+        fp.add_ps_stacker(ps_fg, 'FG', c=psutil.dblue)
+        fp.add_ps_stacker(ps_eor, '21-cm', c=psutil.orange)
+        fp.add_cube(scaled_noise_cube, 'noise', c=psutil.green)
+        fp.done(ncol_legend=4)
+        fp.savefig(os.path.join(plot_out_dir, 'ps_gpr_cmpts.pdf'))
+
+        # PS3D plot
+        fig, ax = plt.subplots()
+        ps3d_res.plot(label='I residual', ax=ax, c=psutil.dblue)
+        ps3d_noise.plot(label='noise I', ax=ax, c=psutil.green)
+        ps3d_rec.plot(label='I residual - noise', ax=ax, c=psutil.orange)
+
+        ax.legend()
+        fig.savefig(os.path.join(plot_out_dir, 'ps3d.pdf'), bbox_inches='tight')
+
+
 @main.command('combine')
 @click.argument('file_list', type=t_file)
 @click.option('--umin', help='Minimum baseline in lambda', type=float, default=10, show_default=True)
@@ -802,258 +1009,31 @@ def combine_sph(file_list, pre_flag, post_flag, output_template, save_intermedia
             c_dt_cube.save(out_dt)
 
 
-class TelescopeSimu(object):
-
-    name = 'none'
-
-    def __init__(self, freqs, dec_deg, hal, har, umin=None, umax=None, timeres=100):
-        self.freqs = freqs
-        self.dec_deg = dec_deg
-        self.hal = hal
-        self.har = har
-        if umin is not None:
-            self.umin = umin
-        if umax is not None:
-            self.umax = umax
-        self.timeres = timeres
-
-    @staticmethod
-    def from_name(instru, freqs, dec_deg, hal, har, umin=None, umax=None, timeres=100):
-        klasses = TelescopeSimu.__subclasses__()
-        [klasses.extend(k.__subclasses__()) for k in klasses[:]]
-
-        for klass in klasses:
-            if hasattr(klass, 'name') and klass.name == instru:
-                return klass(freqs, dec_deg, hal, har, umin, umax, timeres)
-
-        raise ValueError('No instrument with name: %s' % instru)
-
-    def get_stat_pos_file(self):
-        pass
-
-    def get_sefd(self, freq):
-        pass
-
-    def get_primary_beam(self):
-        pass
-
-    def simu_uv(self, include_conj=True):
-        from ps_eor import psutil
-
-        m2a = lambda m: np.squeeze(np.asarray(m))
-
-        lambs = const.c.value / self.freqs
-        umin_meter = (self.umin * lambs).min()
-        umax_meter = (self.umax * lambs).max()
-
-        timev = np.arange(self.hal * 3600, self.har * 3600, self.timeres)
-
-        statpos = np.loadtxt(self.get_stat_pos_file())
-        nstat = statpos.shape[0]
-
-        # All combinations of nant to generate baselines
-        stncom = itertools.combinations(np.arange(1, nstat), 2)
-
-        b1, b2 = zip(*stncom)
-
-        uu = []
-        vv = []
-        ww = []
-
-        print('Simulating UV coverage ...')
-        pr = psutil.progress_report(len(timev))
-        i = 0
-
-        for tt in timev:
-            pr(i)
-            HA = (tt / 3600.) * (15. / 180) * np.pi - (6.8689389 / 180) * np.pi
-            dec = self.dec_deg * (np.pi / 180)
-            RM = np.matrix([[np.sin(HA), np.cos(HA), 0.0],
-                            [-np.sin(dec) * np.cos(HA), np.sin(dec) * np.sin(HA), np.cos(dec)],
-                            [np.cos(dec) * np.cos(HA), - np.cos(dec) * np.sin(HA), np.sin(dec)]])
-            statposuvw = np.dot(RM, statpos.T).T
-            bu = m2a(statposuvw[b1, 0] - statposuvw[b2, 0])
-            bv = m2a(statposuvw[b1, 1] - statposuvw[b2, 1])
-            bw = m2a(statposuvw[b1, 2] - statposuvw[b2, 2])
-
-            ru = np.sqrt(bu ** 2 + bv ** 2)
-            idx = (ru > umin_meter) & (ru < umax_meter)
-
-            uu.extend(bu[idx])
-            vv.extend(bv[idx])
-            ww.extend(bw[idx])
-
-            if include_conj:
-                uu.extend(- bu[idx])
-                vv.extend(- bv[idx])
-                ww.extend(bw[idx])
-
-            i += 1
-
-        return uu, vv, ww
-
-    def simu_gridded_uv(self, uu_meter, vv_meter, fov, oversampling_factor=4):
-        from ps_eor import psutil, datacube
-
-        du = 1 / np.radians(fov)
-        res = 1 / (oversampling_factor * self.umax)
-        n_u = int(np.ceil(1 / (res * du)))
-        shape = (n_u, n_u)
-
-        g_uu, g_vv = psutil.get_uv_grid(shape, res)
-
-        l = [(a - (b - a) / 2., b + (b - a) / 2.) for a, b in psutil.pairwise(g_uu[0])]
-        bins_edges = np.array([k[0] for k in l] + [l[-2][1], l[-1][1]])
-
-        print('Gridding UV coverage ...')
-        weights = []
-        pr = psutil.progress_report(len(self.freqs))
-        for i, lamb in enumerate(const.c.value / self.freqs):
-            pr(i)
-            w, _, _ = np.histogram2d(uu_meter / lamb, vv_meter / lamb, bins=bins_edges)
-            weights.append(w)
-
-        weights = np.array(weights)
-        weights = weights.reshape(len(self.freqs), -1)
-        g_uu = g_uu.flatten()
-        g_vv = g_vv.flatten()
-        ru = np.sqrt(g_uu ** 2 + g_vv ** 2)
-
-        w_min = 0.01 * np.mean(weights[weights > 0])
-
-        idx = (weights.min(axis=0) > w_min) & (ru >= self.umin) & (ru <= self.umax)
-        weights = weights[:, idx]
-        g_uu = g_uu[idx]
-        g_vv = g_vv[idx]
-
-        meta = datacube.ImageMetaData.from_res(res, shape)
-        meta.wcs.wcs.cdelt[2] = psutil.robust_freq_width(self.freqs)
-        meta.set('PEINTTIM', self.timeres)
-        meta.set('PETOTTIM', (self.har - self.hal) * 3600)
-        w_cube = datacube.CartWeightCube(weights, g_uu, g_vv, self.freqs, meta)
-
-        return TelescopGriddedUV(w_cube, self.name, self.get_sefd(self.freqs.mean()))
-
-
-class TelescopGriddedUV(object):
-
-    def __init__(self, weights, name, sefd):
-        from ps_eor import psutil
-
-        self.weights = weights
-        self.name = name
-        self.sefd = sefd
-        self.z = psutil.freq_to_z(self.weights.freqs.mean())
-
-    def save(self, filename):
-        self.weights.meta.set('PESEFD', self.sefd)
-        self.weights.meta.set('PEINSTRU', self.name)
-        self.weights.save(filename)
-
-    @staticmethod
-    def load(filename):
-        from ps_eor import datacube
-
-        weights = datacube.CartWeightCube.load(filename)
-        sefd = weights.meta.get('PESEFD')
-        name = weights.meta.get('PEINSTRU')
-
-        return TelescopGriddedUV(weights, name, sefd)
-
-    def get_ps_gen(self, filter_kpar_min=None, filter_wedge_theta=0):
-        from ps_eor import pspec, datacube
-
-        du = 1 / self.weights.meta.theta_fov
-
-        el = 2 * np.pi * (np.arange(self.weights.ru.min(), self.weights.ru.max(), du))
-
-        ps_conf = pspec.PowerSpectraConfig(el, window_fct='boxcar')
-        ps_conf.filter_kpar_min = filter_kpar_min
-        ps_conf.filter_wedge_theta = filter_wedge_theta
-
-        eor_bin_list = pspec.EorBinList(self.weights.freqs)
-        eor_bin_list.add_freq(1, self.weights.freqs.min() * 1e-6, self.weights.freqs.max() * 1e-6)
-        eor = eor_bin_list.get(1, self.weights.freqs)
-        pb = datacube.PrimaryBeam.from_name(self.name)
-
-        return pspec.PowerSpectraCart(eor, ps_conf, pb)
-
-
-class SkaLow(TelescopeSimu):
-
-    name = 'ska_low'
-    umin = 30
-    umax = 250
-    fov = 4
-
-    def get_stat_pos_file(self):
-        return os.path.join(INSTRU_DIR, 'skalow_enu_statpos.data')
-
-    def get_sefd(self, freq):
-        # Target SEFD at 115 MHz of 2500 Jy
-        return 2500 * (115e6 / freq) ** 0.55
-
-
-class LofarHBA(TelescopeSimu):
-
-    name = 'lofar_hba'
-    umin = 50
-    umax = 250
-    fov = 4
-
-    def get_stat_pos_file(self):
-        return os.path.join(INSTRU_DIR, 'lofar_statpos.data')
-
-    def get_sefd(self, freq):
-        # Typical observed SEFD ~ 140 MHz @ NCP
-        return 4000
-
-
-class NenuFAR(TelescopeSimu):
-
-    name = 'nenufar'
-    umin = 6
-    umax = 60
-    fov = 16
-
-    def get_stat_pos_file(self):
-        return os.path.join(INSTRU_DIR, 'nenufar52_statpos.data')
-
-    def sky_temperature(self, freq, tsys_sky=60):
-        lamb = const.c.value / freq
-        return tsys_sky * lamb ** 2.55
-
-    def inst_temperature(self, freq):
-        """ Instrument temperature at a given frequency ``freq``.
-
-            From: https://github.com/AlanLoh/nenupy/blob/master/nenupy/instru/instru.py
-        """
-        lna_sky = np.array([
-            5.0965, 2.3284, 1.0268, 0.4399, 0.2113, 0.1190, 0.0822, 0.0686,
-            0.0656, 0.0683, 0.0728, 0.0770, 0.0795, 0.0799, 0.0783, 0.0751,
-            0.0710, 0.0667, 0.0629, 0.0610, 0.0614, 0.0630, 0.0651, 0.0672,
-            0.0694, 0.0714, 0.0728, 0.0739, 0.0751, 0.0769, 0.0797, 0.0837,
-            0.0889, 0.0952, 0.1027, 0.1114, 0.1212, 0.1318, 0.1434, 0.1562,
-            0.1700, 0.1841, 0.1971, 0.2072, 0.2135, 0.2168, 0.2175, 0.2159,
-            0.2121, 0.2070, 0.2022, 0.1985, 0.1974, 0.2001, 0.2063, 0.2148,
-            0.2246, 0.2348, 0.2462, 0.2600, 0.2783, 0.3040, 0.3390, 0.3846,
-            0.4425, 0.5167, 0.6183, 0.7689, 1.0086, 1.4042, 2.0732
-        ])
-        freqs = (np.arange(71) + 15) * 1e6
-        return self.sky_temperature(freq) * lna_sky[np.abs(freqs - freq).argmin()]
-
-    def get_sefd(self, freq, tsys_sky=60):
-        d = 5.5
-        lamb = const.c.value / freq
-        tsys = self.sky_temperature(freq, tsys_sky) + self.inst_temperature(freq)
-        a_eff = 19 * np.min([lamb ** 2 / 3., np.ones_like(lamb) * np.pi * d ** 2 / 4.], axis=0)
-
-        return 2 * const.k_B.value / a_eff * 1e26 * tsys
+def isfloat(s):
+    try:
+        float(s)
+    except ValueError:
+        return False
+    else:
+        return True
+
+
+class FloatRangeType(click.ParamType):
+    name = 'float_range'
+
+    def convert(self, value, param, ctx):
+        if isfloat(value):
+            return float(value)
+        if '-' in value:
+            v_s = value.split('-')
+            if len(v_s) == 2 and isfloat(v_s[0]) and isfloat(v_s[1]) and float(v_s[1]) > float(v_s[0]):
+                return (float(v_s[0]), float(v_s[1]))
+        self.fail(f"{value!r} is not a valid float or a range of float", param, ctx)
 
 
 @main.command('simu_uv')
 @click.argument('instru', type=click.Choice(['ska_low', 'nenufar', 'lofar_hba'], case_sensitive=False))
-@click.argument('z', type=float)
+@click.argument('z', type=FloatRangeType())
 @click.option('--out_filename', '-o', help='Base filename of the output file', type=str, default='simu_uv',
               show_default=True)
 @click.option('--umin', help='Minimum baseline', default=None, type=int)
@@ -1064,33 +1044,43 @@ class NenuFAR(TelescopeSimu):
 @click.option('--df', help='Spectral resolution, in kHz', default=195.3, type=float, show_default=True)
 @click.option('--total_time', help='Total observing time (in hours)', default=10, type=float, show_default=True)
 @click.option('--int_time', help='Integration time (in second)', default=100, type=float, show_default=True)
-@click.option('--sefd', help='System Equivalent Flux Density (Jy)', default=None, type=float)
 @click.option('--res', help='Image resolution in degrees. If not given a default 4x oversampling will be used',
               default=None, type=float)
-def simu_uv(instru, z, out_filename, umin, umax, dec_deg, fov, bandwidth, df, total_time, int_time, sefd, res):
-    ''' Compute gridded UV coverage at redshift Z. Possible choices: ska_low, nenufar or lofar_hba.
+# z float or min-max-step
+# freqs range between fmhz_min - bw/2 and fmhz_max +- bw/2
+# output on file for each redshift
+def simu_uv(instru, z, out_filename, umin, umax, dec_deg, fov, bandwidth, df, total_time, int_time, res):
+    ''' Compute gridded UV coverage at redshift Z for instrument INSTRU.
+
+    INSTRU can be any of: ska_low, nenufar or lofar_hba.
+    Z can be a single number or two numbers separated by ','
     '''
-    from ps_eor import psutil
+    from ps_eor import psutil, obssimu
 
     assert total_time < 24, 'total_time of one night must be < 24 hours'
 
-    fmhz = psutil.z_to_freq(z) * 1e-6
-    freqs = np.arange(fmhz - bandwidth / 2, fmhz + bandwidth / 2, df * 1e-3) * 1e6
-    telescop = TelescopeSimu.from_name(instru, freqs, dec_deg, 0, total_time, umin, umax, int_time)
+    if isinstance(z, tuple):
+        fmhz_lower = psutil.z_to_freq(z[1]) * 1e-6
+        fmhz_upper = psutil.z_to_freq(z[0]) * 1e-6
+        freqs = np.arange(fmhz_lower - bandwidth / 2, fmhz_upper + bandwidth / 2, df * 1e-3) * 1e6
+        z_str = f'{z[0]:.1f}-{z[1]:.1f}'
+    else:
+        fmhz = psutil.z_to_freq(z) * 1e-6
+        freqs = np.arange(fmhz - bandwidth / 2, fmhz + bandwidth / 2, df * 1e-3) * 1e6
+        z_str = f'{z:.1f}'
+
+    telescop = obssimu.TelescopeSimu.from_name(instru, freqs, dec_deg, 0, total_time, umin, umax, int_time)
 
     if fov is None:
         fov = telescop.fov
-    if sefd is None:
-        sefd = telescop.get_sefd(freqs.mean())
 
     print(f'Computing gridded UV configuration for instrument {telescop.name}')
     print(f'Simulating {total_time} hour long observation with {int_time} s integration time')
-    print(f'Redshift z = {z:.1f} (mean frequency = {fmhz:.1f} MHz)')
+    print(f'Redshift z = {z_str}')
     print(f'{len(freqs)} frequency channels between {freqs.min() * 1e-6:.1f} and {freqs.max() * 1e-6:.1f} MHz')
     print(f'Frequency resolution: {df:.1f} kHz')
     print(f'Baseline range: {telescop.umin} - {telescop.umax} lambda')
     print(f'FoV: {fov:.1f} degrees')
-    print(f'SEFD: {sefd:.1f} Jy')
 
     uu_meter, vv_meter, _ = telescop.simu_uv()
 
@@ -1099,12 +1089,12 @@ def simu_uv(instru, z, out_filename, umin, umax, dec_deg, fov, bandwidth, df, to
     else:
         osamp_factor = 4
 
-    uv_simu = telescop.simu_gridded_uv(uu_meter, vv_meter, fov, osamp_factor)
+    uv_simu = telescop.simu_gridded_uv(uu_meter, vv_meter, fov, osamp_factor, min_weight=0)
 
     print(f'Image size: {uv_simu.weights.meta.shape}')
     print(f'Image resolution: {np.degrees(uv_simu.weights.meta.res) * 60:.2f} arcmin (oversampling: {osamp_factor})')
 
-    name = f'{out_filename}_{telescop.name}_z{z:.1f}_{total_time}h'
+    name = f'{out_filename}_{telescop.name}_z{z_str}_{total_time}h'
 
     fig, ax = plt.subplots(figsize=(5, 4))
     uv_simu.weights.plot_uv(uv_lines=[], ax=ax)
@@ -1122,17 +1112,23 @@ def simu_uv(instru, z, out_filename, umin, umax, dec_deg, fov, bandwidth, df, to
 @click.argument('uv_simu_h5', type=t_file)
 @click.argument('total_time_hour', type=int)
 @click.option('--out_filename', '-o', help='Base filename of the output file', type=str, default='simu_noise')
-def simu_noise_img(uv_simu_h5, total_time_hour, out_filename):
+@click.option('--min_weight', help='Minimum number of visibilities in a uv-cell', default=10, type=int, show_default=True)
+def simu_noise_img(uv_simu_h5, total_time_hour, out_filename, min_weight):
     ''' Compute noise fits cube from a simulated UV coverage UV_SIMU_H5 for TOTAL_TIME_HOUR observation. '''
 
-    from ps_eor import datacube, psutil
+    from ps_eor import datacube, psutil, obssimu
 
-    uv_simu = TelescopGriddedUV.load(uv_simu_h5)
+    uv_simu = obssimu.TelescopGriddedUV.load(uv_simu_h5)
+    freqs = uv_simu.weights.freqs
 
-    print(f'Simulating thermal noise with SEFD = {uv_simu.sefd} Jy ...')
-    noise_cube = uv_simu.weights.simulate_noise(uv_simu.sefd, total_time_hour * 3600, hermitian=False)
+    print(uv_simu.weights.ru.min(), uv_simu.weights.ru.max(), np.diff(uv_simu.weights.uu)[0])
+
+    print(f'Simulating thermal noise with mean SEFD = {uv_simu.get_sefd().mean()} Jy ...')
+    print(f'{len(freqs)} frequency channels between {freqs.min() * 1e-6:.1f} and {freqs.max() * 1e-6:.1f} MHz')
+    noise_cube = uv_simu.weights.simulate_noise(uv_simu.get_sefd(), total_time_hour * 3600, hermitian=False)
+    noise_cube.filter_min_weight(min_weight, replace=True)
     noise_img = noise_cube.image()
-    print(f'RMS noise @ mid frequency {noise_img.data[len(noise_img.freqs) // 2].std()} Kelvin')
+    print(f'RMS noise @ mid frequency {noise_img.data[len(freqs) // 2].std()} Kelvin')
 
     name = f'{out_filename}_{uv_simu.name}_z{uv_simu.z:.1f}_{total_time_hour}h.fits'
     print(f'Saving noise cube to {name} ...')
@@ -1143,50 +1139,136 @@ def simu_noise_img(uv_simu_h5, total_time_hour, out_filename):
 @click.argument('uv_simu_h5', type=t_file)
 @click.argument('total_time_hour', type=float)
 @click.option('--out_filename', '-o', help='Base filename of the output file', type=str, default='simu_noise')
-@click.option('--n_simu', help='Number of simulation to perform', default=10, type=int)
-@click.option('--kmax', help='Maximum k-mode', default=0.5, type=float)
-@click.option('--nks', help='Number of modes', default=7, type=int)
+@click.option('--kmin', help='Minimum k-mode. Default is the minimum k-mode accessible', default=None,
+              type=float, show_default=False)
+@click.option('--kmax', help='Maximum k-mode', default=0.5, type=float, show_default=True)
+@click.option('--nks', help='Number of modes', default=7, type=int, show_default=True)
 @click.option('--filter_kpar_min', help='Filter mode below kpar_min', default=None, type=float)
 @click.option('--filter_wedge_theta', help='Filter mode below the wedge line', default=0, type=float)
-def simu_noise_ps(uv_simu_h5, total_time_hour, out_filename, n_simu, kmax, nks, filter_kpar_min, filter_wedge_theta):
+@click.option('--sefd', help='System Equivalent Flux Density (Jy)', default=None, type=float)
+@click.option('--min_weight', help='Minimum number of visibilities in a uv-cell', default=10, type=int, show_default=True)
+def simu_noise_ps(uv_simu_h5, total_time_hour, out_filename, kmin, kmax, nks, 
+                  filter_kpar_min, filter_wedge_theta, sefd, min_weight):
     ''' Compute noise PS from a simulated UV coverage UV_SIMU_H5 for TOTAL_TIME_HOUR observation. '''
 
-    from ps_eor import datacube, psutil
+    from ps_eor import datacube, psutil, obssimu
 
-    uv_simu = TelescopGriddedUV.load(uv_simu_h5)
+    uv_simu = obssimu.TelescopGriddedUV.load(uv_simu_h5)
+    freqs = uv_simu.weights.freqs
     ps_gen = uv_simu.get_ps_gen(filter_kpar_min, filter_wedge_theta)
-    kbins = np.logspace(np.log10(ps_gen.kmin), np.log10(kmax), nks)
 
-    print('Simulating noise cube ...')
-    pss = {'ps': [], 'ps2d': [], 'ps3d': []}
-    pr = psutil.progress_report(n_simu)
-    for i in range(n_simu):
-        pr(i)
-        noise_cube = uv_simu.weights.simulate_noise(uv_simu.sefd, total_time_hour * 3600, hermitian=False)
-        pss['ps'].append(ps_gen.get_ps(noise_cube))
-        pss['ps2d'].append(ps_gen.get_ps2d(noise_cube))
-        pss['ps3d'].append(ps_gen.get_ps3d(kbins, noise_cube))
+    if kmin is None:
+        kmin = ps_gen.kmin
+    kbins = np.logspace(np.log10(kmin), np.log10(kmax), nks)
+
+    if sefd is None:
+        sefd = uv_simu.get_sefd()
+        print(f'Simulating thermal noise with mean SEFD = {sefd.mean():.1f} Jy ...')
+    else:
+        print(f'Simulating thermal noise with SEFD = {sefd:.1f} Jy ...')
+
+    print(f'{len(freqs)} frequency channels between {freqs.min() * 1e-6:.1f} and {freqs.max() * 1e-6:.1f} MHz')
+    noise_std = uv_simu.get_noise_std_cube(total_time_hour * 3600, sefd=sefd, min_weight=min_weight)
+    pss = {}
+    pss['ps'] = ps_gen.get_ps(noise_std)
+    pss['ps2d'] = ps_gen.get_ps2d(noise_std)
+    pss['ps3d'] = ps_gen.get_ps3d(kbins, noise_std)
 
     print('Generating power-spectra ...')
 
     for ps_name, ps in pss.items():
-        ps_mean = ps[0].new_with_data(np.mean([k.data for k in ps], axis=0), np.mean([k.err for k in ps], axis=0))
-
         name = f'{out_filename}_{ps_name}_{uv_simu.name}_z{uv_simu.z:.1f}_{total_time_hour}h'
 
         print(f'Saving {ps_name} to {name} ...')
 
         fig, ax = plt.subplots(figsize=(5, 4))
-        ps_mean.plot(ax, label='Noise')
+        ps.plot(ax, label='Noise')
 
         if ps_name == 'ps3d':
-            ax.plot(ps_mean.k_mean, 2 * ps_mean.err * 1e6, label='2-sigma sensitivity')
+            ax.plot(ps.k_mean, 2 * ps.err * 1e6, label='2-sigma sensitivity')
             ax.legend()
 
         fig.tight_layout()
         fig.savefig(f'{name}.pdf')
 
-        ps_mean.save_to_txt(f'{name}.txt')
+        ps.save_to_txt(f'{name}.txt')
+
+
+@main.command('simu_noise_ps_zrange')
+@click.argument('uv_simu_h5', type=t_file)
+@click.argument('total_time_hour', type=float)
+@click.argument('k', type=float)
+@click.option('--out_filename', '-o', help='Base filename of the output file', type=str, default='simu_noise')
+@click.option('--bandwidth', help='Bandwidth, in MHz', default=51 * 195.3 * 1e-3, type=float, show_default=True)
+@click.option('--delta_z', help='Redshift step', default=0.5, type=float, show_default=True)
+@click.option('--dk_over_k', help='k bin size in dk/k', default=0.4, type=float, show_default=True)
+@click.option('--filter_kpar_min', help='Filter mode below kpar_min', default=None, type=float)
+@click.option('--filter_wedge_theta', help='Filter mode below the wedge line', default=0, type=float)
+@click.option('--min_weight', help='Minimum number of visibilities in a uv-cell', default=10, type=int, show_default=True)
+@click.option('--smooth', help='Smooth the resulting sensitivity curve', is_flag=True)
+# allow for multiple uv_simu_h5
+# add k_min
+def simu_noise_ps_zrange(uv_simu_h5, total_time_hour, k, out_filename, bandwidth, delta_z, dk_over_k,
+                         filter_kpar_min, filter_wedge_theta, min_weight, smooth):
+    ''' Compute noise PS from a simulated UV coverage UV_SIMU_H5 for TOTAL_TIME_HOUR observation at k of K.'''
+    from ps_eor import psutil, obssimu
+    from scipy.signal import savgol_filter
+
+    uv_simu = obssimu.TelescopGriddedUV.load(uv_simu_h5)
+    z_max = np.round(psutil.freq_to_z(uv_simu.weights.freqs.min() + bandwidth / 2 * 1e6), 1)
+    z_min = np.round(psutil.freq_to_z(uv_simu.weights.freqs.max() - bandwidth / 2 * 1e6), 1)
+    zs = np.arange(z_min, z_max + delta_z, delta_z)
+
+    k_bin = [(1 - dk_over_k / 2) * k, (1 + dk_over_k / 2) * k]
+
+    print(f'Generating PS at k = {k}, k_bin = {k_bin[0]:.3f} - {k_bin[1]:.3f}, z = {z_min} - {z_max}')
+
+    pks = []
+    pks_err = []
+
+    for z in zs:
+        mfreq = psutil.z_to_freq(z)
+        uv_simu_z = uv_simu.get_slice(mfreq - bandwidth / 2 * 1e6, mfreq + bandwidth / 2 * 1e6)
+
+        freqs = uv_simu_z.weights.freqs
+        ps_gen = uv_simu_z.get_ps_gen(filter_kpar_min, filter_wedge_theta)
+
+        print(f'Redshift z = {z}')
+        print(f'Simulating thermal noise with mean SEFD = {uv_simu_z.get_sefd().mean():.1f} Jy ...')
+        print(f'{len(freqs)} frequency channels between {freqs.min() * 1e-6:.1f} and {freqs.max() * 1e-6:.1f} MHz')
+
+        noise_std = uv_simu_z.get_noise_std_cube(total_time_hour * 3600, min_weight=min_weight)
+
+        ps3d = ps_gen.get_ps3d(k_bin, noise_std)
+        pks.append(ps3d.data[0] * 1e6)
+        pks_err.append(ps3d.err[0] * 1e6)
+
+    pks_err = np.array(pks_err)
+    pks = np.array(pks)
+
+    if smooth:
+        w_l = min(int(psutil.get_next_odd(2 / delta_z)), int(psutil.get_previous_odd(len(zs))))
+        print(w_l)
+        pks = 10 ** savgol_filter(np.log10(pks), w_l, 1)
+        pks_err = 10 ** savgol_filter(np.log10(pks_err), w_l, 1)
+
+    array_data = np.array([zs, [k_bin[0]] * len(zs), [k_bin[1]] * len(zs), [k] * len(zs), pks, pks_err]).T
+    header = f'Spherically averaged power-spectra {uv_simu.name} {total_time_hour}h\n'
+    header += f'Parameters: BW={bandwidth:.1f} MHz, kpar_min={filter_kpar_min}, filter_wedge={filter_wedge_theta} deg, dk/k={dk_over_k}\n'
+    header += (r'z, k_min (h cMpc^-1), k_max (h cMpc^-1), k_mean (h cMpc^-1), \Delta^2 (mK^2), \Delta_err^2 (mK^2)')
+
+    name = f'{out_filename}_ps_{uv_simu.name}_z{z_min:.1f}-{z_max:.1f}_{total_time_hour}h_k{k}'
+    np.savetxt(f'{name}.txt', array_data, fmt='%14.4f', header=header, delimiter=' ')
+
+    fig, ax = plt.subplots(figsize=(5, 4))
+    ax.plot(zs, pks, label=f'Noise at k={k}')
+    ax.plot(zs, 2 * pks_err, label=f'2-sigma sensitivity at k={k}')
+    ax.set_yscale('log')
+    ax.set_xlabel('Redshift')
+    ax.set_ylabel(r'$\Delta^2 (k)\,[\mathrm{mK^2}]$')
+    ax.legend()
+    fig.tight_layout()
+    fig.savefig(f'{name}.pdf')
 
 
 if __name__ == '__main__':
